@@ -10,6 +10,8 @@ Host code in the understory.
 # TODO PEP 592 -- Adding "Yank" Support to the Simple API
 # TODO PEP 658 -- Serve Distribution Metadata in the Simple Repository API
 
+import re
+import subprocess
 from pathlib import Path
 
 import warez
@@ -18,10 +20,16 @@ import web
 app = web.application(
     __name__,
     prefix="code",
-    args={"project": r"[a-z0-9.-]+", "package": r"[\w.-]+"},
+    args={
+        "project": r"[a-z0-9.-]+",
+        "release": r"((\d+\.)?\d+\.)?\d+",
+        "filename": r"[\w./\-]+",
+        "package": r"[\w.-]+",
+    },
     model={
         "projects": {
             "name": "TEXT UNIQUE",
+            "pypi": "TEXT UNIQUE",
         },
         "packages": {
             "project_id": "INTEGER",
@@ -42,14 +50,13 @@ app = web.application(
     },
 )
 
-project_dir = Path("projects")
-package_dir = Path("packages")
+code_dir = Path("code")
 
 
 @app.query
 def create_project(db, name):
     """Create a project."""
-    return db.insert("projects", name=name)
+    return db.insert("projects", name=name, pypi=name)
 
 
 @app.query
@@ -58,15 +65,20 @@ def get_projects(db):
     return [r["name"] for r in db.select("projects", what="name", order="name")]
 
 
+@app.query
+def get_project_from_pypi(db, pypi_name):
+    """Return the project associated with `pypi_name`."""
+    return db.select("projects", what="name", where="pypi = ?", vals=[pypi_name])[0][
+        "name"
+    ]
+
+
 @app.query
 def create_package(db, form):
     """Create a project."""
-    try:
-        project_id = db.insert("projects", name=form.name)
-    except db.IntegrityError:
-        project_id = db.select(
-            "projects", what="rowid, name", where="name = ?", vals=[form.name]
-        )[0]["rowid"]
+    project_id = db.select(
+        "projects", what="rowid, name", where="name = ?", vals=[form.name]
+    )[0]["rowid"]
     return db.insert(
         "packages",
         project_id=project_id,
@@ -97,6 +109,22 @@ def get_packages(db, project):
     )
 
 
+@app.query
+def get_package_versions(db, project):
+    """Return a list of packages for given project."""
+    return [
+        r["version"]
+        for r in db.select(
+            "packages",
+            what="DISTINCT version",
+            join="""projects ON packages.project_id = projects.rowid""",
+            where="projects.name = ?",
+            vals=[project],
+            order="version",
+        )
+    ]
+
+
 @app.control("")
 class Code:
     """Code index."""
@@ -109,11 +137,16 @@ class Code:
 
     def post(self):
         """Create a project."""
-        name = web.form("name").name
-        app.model.create_project(name)
-        project_dir.mkdir(exist_ok=True)
-        warez.Repo(project_dir / name, init=True)
-        return web.Created(app.view.project_created(name), f"/{name}")
+        project = web.form("project").project
+        app.model.create_project(project)
+        project_dir = code_dir / project
+        bare_repo = project_dir / "source.git"
+        working_repo = project_dir / "working"
+        repo = warez.Repo(bare_repo, init=True, bare=True)
+        repo.update_server_info()
+        repo.config("http.receivepack", "true")
+        warez.clone_repo(bare_repo, working_repo)  # TODO shallow?
+        return web.Created(app.view.project_created(project), f"/{project}")
 
 
 @app.control("{project}")
@@ -124,25 +157,59 @@ class Project:
         """Return details about the project."""
         return app.view.project(
             project,
-            warez.Repo(project_dir / project),
-            app.model.get_packages(project),
+            warez.Repo(code_dir / project / "source.git"),
+            app.model.get_package_versions(project),
         )
 
 
-@app.control("{project}/packages/{package}")
-class Package:
-    """Project package."""
+@app.control("{project}.git")
+class ProjectRedirect:
+    """Project .git redirect."""
 
-    def get(self, project, package):
+    def get(self, project):
+        """Redirect to main project index."""
+        raise web.SeeOther(project)
+
+
+@app.control("{project}/releases/{release}")
+class ProjectRelease:
+    """A project's release."""
+
+    def get(self, project, release):
         """Return the package file."""
-        return package_dir / package
+        files = sorted((code_dir / project / "releases" / release).iterdir())
+        return app.view.project_release(project, release, files)
+
+
+@app.control("{project}/releases/{release}/files/{filename}")
+class ProjectReleaseFile:
+    """A file in a project's release."""
+
+    def get(self, project, release, filename):
+        """Return the release's file."""
+        filepath = code_dir / project / "releases" / release / filename
+        try:
+            with filepath.open() as fp:
+                content = fp.read()
+        except IsADirectoryError:
+            content = filepath.iterdir()
+        return app.view.project_release_file(project, release, filename, content)
+
+
+@app.control("{project}/releases/{release}")
+class ProjectReleasePackage:
+    """A project's release package."""
+
+    def get(self, project, release):
+        """Return the package file."""
+        return code_dir / project / "releases" / release
 
 
 @app.control("_pypi")
 class PyPIIndex:
     """PyPI repository in Simple Repository format."""
 
-    owner_only = ["post"]
+    # TODO owner_only = ["post"]
 
     def get(self):
         """Return a simplified list of the repository's projects."""
@@ -153,8 +220,28 @@ class PyPIIndex:
         form = web.form(":action")
         if form[":action"] != "file_upload":
             raise web.BadRequest(f"Provided `:action={form[':action']}` not supported.")
+        release_file = form.content.save(file_dir="/tmp")
+        release_name, release_remaining = re.match(
+            r"([\w-]+)\-(\d+\.\d+\.\d+.*)", release_file.name
+        ).groups()
+        project = app.model.get_project_from_pypi(release_name)
+        releases_dir = code_dir / project / "releases"
+        releases_dir.mkdir(exist_ok=True)
+        release_file = release_file.replace(releases_dir / release_remaining)
+        if release_file.suffix == ".gz":
+            release_version = release_file.stem[:-4]
+            release_dir = f"{release_name}-{release_version}"
+            subprocess.run(
+                [
+                    "tar",
+                    "xf",
+                    release_file.name,
+                    releases_dir / release_dir,
+                ],
+                cwd=releases_dir,
+            )
+            (releases_dir / release_dir).replace(releases_dir / release_version)
         app.model.create_package(form)
-        form.content.save(file_dir=package_dir)
         raise web.Created(
             "Package has been uploaded.",
             "/{form.name}/packages/{form.content.fileobj.filename}",
